build-lists: true
autoscale: true

# Programming with nothing

## An *(brief)* Intro to the _`λ`-calculus_

---

![](http://media3.giphy.com/media/yoJC2A59OCZHs1LXvW/giphy.gif)

---

![right](https://pbs.twimg.com/profile_images/787093618871898112/LaOLzgCs.jpg)

I'm *Professor* **Michael Selsky**

iOS Developer on MNE NextGen

[@MichaelSelsky](https://twitter.com/michaelselsky) 

---

# `λ`-Calculus

---

![Giphy](http://media2.giphy.com/media/l2YWF00ZX8wOs0p0s/giphy.gif)

#[fit] `λ`-calculus ≠ ∫ calculus

^ here calculus just means a system of calculation/computation. Because Math names are weird

---

# `λ`-Calculus

- Turing complete
- Based entirely around functions
- Invented in the 1930s, before the Turing Machine



---

# Alonzo Church

- Invented it to try to solve the Entscheidungsproblem
- Published a paper proving it to be impossible

![right](https://upload.wikimedia.org/wikipedia/en/a/a6/Alonzo_Church.jpg)

---

While λ-calculus is really cool, it can be difficult to read so instead we'll use

---

![](https://media4.giphy.com/media/ROyP6SWlAv2Za/giphy.gif)

---


![fit](https://ee5817f8e2e9a2e34042-3365e7f0719651e5b8d0979bce83c558.ssl.cf5.rackcdn.com/python.png)

---

# Disclaimers
1. This is *not* software engineering advice
2. Don't do this in production
3. No *seriously*, I'll reject that PR
4. This is a learning *exercise* and a *game*
5. I am *not* an expert at this


---

![Giphy](http://media4.giphy.com/media/3o7aTkjnoAzNaxPes8/giphy.gif)

# [fit] Game means **Rules**

---

# **Rules**
1. You can create functions
2. You can call functions
3. Functions can only take one argument and return one value
4. Can't use the `def` keyword
5. That's it

---

# [fit] That's it

---


```python

def fact(n):
    if n == 0:
        return 1
    else:
        return n * fact(n-1)

print(fact(6))

```

---

# 720

---

```python

def fact(n):
    if n == 0:
        return 1
    else:
        return n * fact(n-1)

print(fact(6))

```

---

```python

ONE = 1

def fact(n):
    if n == 0:
        return ONE
    else:
        return n * fact(n-1)

print(fact(6))

```

---

```python

ONE = 1
IS_ZERO = lambda x: x == 0

def fact(n):
    if IS_ZERO(n):
        return ONE
    else:
        return n * fact(n-1)

print(fact(6))

```

---

` fst = λx. λy. x `
` snd = λx. λy. y `

---

```python

ONE = 1
IS_ZERO = lambda x: x == 0

def fact(n):
    if IS_ZERO(n):
        return ONE
    else:
        return n * fact(n-1)

print(fact(6))

```

---

```python

ONE = 1
IS_ZERO = lambda x: x == 0
SUB_1 = lambda x: x - 1

def fact(n):
    if IS_ZERO(n):
        return ONE
    else:
        return n * fact(SUB_1(n))

print(fact(6))

```

---

```python

ONE = 1
IS_ZERO = lambda x: x == 0
SUB_1 = lambda x: x - 1
MULT = lambda x, y: x * y

def fact(n):
    if IS_ZERO(n):
        return ONE
    else:
        return MULT(n, fact(SUB_1(n)))

print(fact(6))

```

---

![Giphy](https://media0.giphy.com/media/xT8qBqNisx9dkXWAX6/giphy.gif)   

# 720

---

```python

ONE = 1
IS_ZERO = lambda x: x == 0
SUB_1 = lambda x: x - 1
MULT = lambda x, y: x * y

def fact(n):
    if IS_ZERO(n):
        return ONE
    else:
        return MULT(n, fact(SUB_1(n)))

print(fact(6))

```

---

```python

ONE = 1
IS_ZERO = lambda x: x == 0
SUB_1 = lambda x: x - 1
MULT = lambda x, y: x * y
IF = lambda cond, t_val, f_val: t_val if cond else f_val

def fact(n):
    return IF(IS_ZERO(n),
              ONE,
              MULT(n, fact(SUB_1(n))))

print(fact(6))

```

---

# [fit]​(ノಥ,_｣ಥ)ノ彡┻━┻
![Giphy](https://media1.giphy.com/media/BKjI4ysCCzEwo/giphy.gif)

---

![Giphy](http://media1.giphy.com/media/xlTwaFb20TVjW/giphy.gif)

## RecursionError: maximum recursion depth exceeded in comparison

---

```python

ONE = 1
IS_ZERO = lambda x: x == 0
SUB_1 = lambda x: x - 1
MULT = lambda x, y: x * y
IF = lambda cond, t_val, f_val: t_val if cond else f_val

def fact(n):
    return IF(IS_ZERO(n),
              ONE,
              MULT(n, fact(SUB_1(n))))

print(fact(6))

```

---

![Giphy](http://media4.giphy.com/media/l2YWmdzX0DkNHk3ew/giphy.gif)

But, we can *fix* this

---

```python

ONE = 1
IS_ZERO = lambda x: x == 0
SUB_1 = lambda x: x - 1
MULT = lambda x, y: x * y
IF = lambda cond, t_func, f_func: t_func(None) if cond else f_func(None)

def fact(n):
    return IF(IS_ZERO(n),
              lambda _: ONE,
              lambda _: MULT(n, fact(SUB_1(n))))

print(fact(6))

```

---



#  720

## [fit] ┬──┬ ノ( ゜-゜ノ)

---

```python

ONE = 1
IS_ZERO = lambda x: x == 0
SUB_1 = lambda x: x - 1
MULT = lambda x, y: x * y
IF = lambda cond, t_func, f_func: t_func(None) if cond else f_func(None)

def fact(n):
    return IF(IS_ZERO(n),
              lambda _: ONE,
              lambda _: MULT(n, fact(SUB_1(n))))

print(fact(6))

```

---

![filtered](http://media2.giphy.com/media/D1Dgzvbbout4A/giphy.gif)

# Let's get rid of that *def*

---

```python

fact = lambda n:
          IF(IS_ZERO(n),
              lambda _: ONE,
              lambda _: MULT(n, fact(SUB_1(n))))

print(fact(6))

```

---

Except we can't use a name in it's own definition.</br>That's just weird.

---

So we can just pass it in instead

---

```python

fact =  lambda fact, n:
          IF(IS_ZERO(n),
              lambda _: ONE,
              lambda _: MULT(n, fact(fact, SUB_1(n))))

print(fact(fact, 6))

```

---

```python

fact =  lambda myself, n:
            IF(IS_ZERO(n),
              lambda _: ONE,
              lambda _: MULT(n, myself(myself, SUB_1(n))))

print(fact(fact, 6))

```

---

# Remember **Rule 3**

---

# **Rule 3**

### Functions can only take *one* argument

---

```python

MULT = lambda x, y: x * y
IF = lambda cond, t_func, f_func: t_func(None) if cond else f_func(None)
fact =  lambda myself, n: ...

```

These take 2 or 3 arguments

---

![Giphy](http://media1.giphy.com/media/qW0cAj6K9PqdW/giphy.gif)

But, we can *fix* this

---

^ But first we need to talk about

# Higher Order Functions

---

Functions that take and/or return other functions

---

![Giphy](http://media2.giphy.com/media/ohdY5OaQmUmVW/giphy.gif)

# [fit] Math

---

$$
\lim_{x\to\infty} f(x)
$$

---

$$
\int_{a}^{b} x^2 dx
$$

---

$$
\sum_{n=1}^{\infty} 2^{-n}
$$

---

![Giphy](http://media4.giphy.com/media/EDyM2g9vX2BXy/giphy.gif)

# [fit] Functional Programming

---

```Swift
[1, 2, 3, 4].map(square) # => [1, 4, 9, 16]
```

---

```Swift
[1, 2, 3, 4].filter(isEven) # => [2, 4]
```

---

```Swift

[1, 2, 3, 4].reduce(0, +) # => 10

```

---

![Giphy](http://media0.giphy.com/media/PcXsCu5ybj2Ra/giphy.gif)

# [fit] Currying

---

# [fit] Schönfinkelisation

---

> Technique of translating the evaluation of a function that takes multiple arguments into evaluating a sequence of functions, each with a single argument.
-- Wikipedia

---

# Haskell Curry
- Traced back to 1893
- Curry-Howard correspondence 

![right](https://wiki.haskell.org/wikiupload/8/86/HaskellBCurry.jpg)

---

```python

add = lambda x, y = x + y
add(2, 3) # => 5

add = lambda x: lambda y: x + y
add(2)(3) # => 5

add2 = add(2)
add2(6) # => 8

```

---

```python
ONE = 1
IS_ZERO = lambda x: x == 0
SUB_1 = lambda x: x - 1
MULT = lambda x: lambda y: x * y
IF = lambda cond: lambda t_func: lambda f_func: f_func: t_func(None) if cond else f_func(None)

fact =  lambda myself: lambda n:
      IF(
        IS_ZERO(n)
      )(
        lambda _: ONE
      )(
        lambda _: MULT(n)(myself(myself)(SUB_1(n)))
      )

print(fact(fact)(6))

```

---

# Cool, so *everything* is a function

---

# Except

---

``` python

ONE = 1

```

---

![Giphy](http://media2.giphy.com/media/PrEUkNFD9pN2o/giphy.gif)

But, we can *fix* this

---

![filtered](http://media4.giphy.com/media/l0ExncehJzexFpRHq/giphy.gif)
![filtered](http://media0.giphy.com/media/26gsqQxPQXHBiBEUU/giphy.gif)
![filtered](http://media1.giphy.com/media/l0EwYkyU1JCExVquc/giphy.gif)
![filtered](http://media2.giphy.com/media/d1E1szXDsHUs3WvK/giphy.gif)
# [fit] **Numbers**

---

*Alonzo Church* figured this out for us

![left](https://upload.wikimedia.org/wikipedia/en/a/a6/Alonzo_Church.jpg)


----

# Church Numerals

![](https://images-na.ssl-images-amazon.com/images/I/41l15I7I9ML.jpg)

---

#Church Numerals

- Just another way to represent positive integers
- Think *"Roman Numerals"* but with functions
- Work by applying the same function over and over

---

A number `n` is defined as the function `f` applied `n` times to a based value of `x`

---

```

0 = x
1 = f(x)
2 = f(f(x))
3 = f(f(f(x)))
...
n = f(f(f(... n times ...(x))))
```

---

```python

ZERO = lambda f: lambda x: x
ONE = lambda f: lambda x: f(x)
TWO = lambda f: lambda x: f(f(x))

```

---

# Conversion

---

```python
TWO(lambda x: x + 1)(0) # => 2
```

---

```python
THREE(lambda: x * 2)(1) # => 8
```

---

# Addition

---

So let's build this up through a few simple cases

---

## Simplest Case

---

### Adding 0<br />AKA Identity

---


```python

IDENTITY = lambda n: lambda f: lambda x: n(f)(x)

```

- Take in a number `n` and return a function that takes in a `f` and `x` and then passes those into the number `n`

- We can think of any function that takes in a `f` and `x` as a number

---

## Simple Case

---

### Add 1<br />AKA Successor

---

Remember what how we define a number

---

A number `n` is defined as the function `f` applied `n` times to a based value of `x`

---

So adding one is just calling `f` one more time


---

```python

SUCC = lambda n: lambda f: lambda x: f(n(f)(x))

```

- Take in a number `n` and return another number
  - `lambda f: lambda x:`
- Apply `f` one more time to number `n`

---

``` python

TWO = lambda f: lambda x: f(f(x))

SUCC = lambda n: lambda f: lambda x: f(n(f)(x))

THREE = SUCC(TWO) # => f(f(f(n)))

```

---

# Addition

---

Remember what how we define a number

---

A number `n` is defined as the function `f` applied `n` times to a based value of `x`

---

```

m + n = m(SUCC)(n)

```

We apply add 1 `m` times to `n`

```python
ADD = lambda n: lambda m: n(SUCC)(m)
```

---

We're also going to need *predecessor* aka `SUB_1`

---

```python
PRED = lambda n:
  lambda f: lambda x:
    n(lambda g: lambda h: h(g(f)))
    (lambda u: x)(lambda u: u)
```

This applies `f` one less time.

It's a little too confusing to explain.

---

``` python

MULT = (
      lambda n: lambda m: n(lambda x:
                            ADD(x)(m)) (ZERO)
  )

```

---

# And that's **Numbers**

![](http://www.drodd.com/images15/numbers6.jpg)

---

That will be on the quiz

---

That will be on the quiz

😁

---

Another data type

---

# [fit] Booleans

---

So what are booleans?

---

![fit](http://media4.giphy.com/media/sf9usCEvitius/giphy.gif)
# [fit] Booleans = Choice

---

``` Python
NIL = lambda x: x
TRUE = lambda t_func: lambda f_func: t_func(NIL)
FALSE = lambda t_func: lambda f_func: f_func(NIL)

IF = lambda cond: lambda t_func: lambda f_func:
    cond(t_func)(f_func)

IS_ZERO = lambda n: n(lambda x: FALSE)(TRUE)

```

---

![Giphy](http://media0.giphy.com/media/vv41HlvfogHAY/giphy.gif)
# Now we have all the pieces to do this

---

``` python

NULL = lambda x: x
TRUE = lambda t: lambda f: t(NULL)
FALSE = lambda t: lambda f: f(NULL)
IF = lambda cond: lambda t: lambda f: cond(t)(f)

ZERO = lambda f: lambda x: x
ADD1 = lambda n: (lambda f: lambda x: f(n (f)(x)))
ADD = lambda n: lambda m: n(ADD1)(m)
ONE = ADD1(ZERO)
IS_ZERO = lambda n: n(lambda x: FALSE)(TRUE)
SUB1 = (
  (lambda n:
   lambda f:
   lambda x: n (lambda g: lambda h: h(g(f)))
               (lambda u: x)
               (lambda u: u))
)
MULT = (
  lambda n: lambda m: n(lambda x:
                        ADD(x)(m)) (ZERO)
)

SIX = ADD1(ADD1(ADD1(ADD1(ADD1(ADD1(ZERO))))))

```

---

```python

FACT = (
          lambda myself: (
              lambda n: (
                  IF(
                      IS_ZERO(n)
                  )(
                      lambda _: ONE
                  )(
                      lambda _: MULT(n)(myself(myself)(SUB1(n)))
                  )
              )
          )
      )

print(FACT(FACT)(SIX))

```

---

`<function __main__.<lambda>.<locals>.<lambda>>`

---

Cool it *works*

---

```python

print(FACT(FACT)(SIX)(lambda x: x + 1)(0))

```

---

# [fit] 720

---

```python

def fact(n):
    if n == 0:
        return 1
    else:
        return n * fact(n-1)

print(fact(6))

```

---

```python

NULL = lambda x: x
TRUE = lambda t: lambda f: t(NULL)
FALSE = lambda t: lambda f: f(NULL)
IF = lambda cond: lambda t: lambda f: cond(t)(f)

ZERO = lambda f: lambda x: x
ADD1 = lambda n: (lambda f: lambda x: f(n (f)(x)))
ADD = lambda n: lambda m: n(ADD1)(m)
ONE = ADD1(ZERO)
IS_ZERO = lambda n: n(lambda x: FALSE)(TRUE)
SUB1 = (
  (lambda n:
   lambda f:
   lambda x: n (lambda g: lambda h: h(g(f)))
               (lambda u: x)
               (lambda u: u))
)
MULT = (
  lambda n: lambda m: n(lambda x:
                        ADD(x)(m)) (ZERO)
)

SIX = ADD1(ADD1(ADD1(ADD1(ADD1(ADD1(ZERO))))))

FACT = (
          lambda myself: (
              lambda n: (
                  IF(
                      IS_ZERO(n)
                  )(
                      lambda _: ONE
                  )(
                      lambda _: MULT(n)(myself(myself)(SUB1(n)))
                  )
              )
          )
      )

print(FACT(FACT)(SIX))

```

---

So we're *done*

---

# [fit] Except

---

^ Look at all those assignments. That's not allowed in the rules.

```python

NULL = lambda x: x
TRUE = lambda t: lambda f: t(NULL)
FALSE = lambda t: lambda f: f(NULL)
IF = lambda cond: lambda t: lambda f: cond(t)(f)

ZERO = lambda f: lambda x: x
ADD1 = lambda n: (lambda f: lambda x: f(n (f)(x)))
ADD = lambda n: lambda m: n(ADD1)(m)
ONE = ADD1(ZERO)
IS_ZERO = lambda n: n(lambda x: FALSE)(TRUE)
SUB1 = (
  (lambda n:
   lambda f:
   lambda x: n (lambda g: lambda h: h(g(f)))
               (lambda u: x)
               (lambda u: u))
)
MULT = (
  lambda n: lambda m: n(lambda x:
                        ADD(x)(m)) (ZERO)
)

SIX = ADD1(ADD1(ADD1(ADD1(ADD1(ADD1(ZERO))))))

FACT = (

```

---

![Giphy](https://media4.giphy.com/media/TKLty6gdiwaXe/giphy.gif)

But, we can *fix* this

---

```python

NULL = lambda x: x
TRUE = lambda t: lambda f: t(NULL)
FALSE = lambda t: lambda f: f(NULL)
IF = lambda cond: lambda t: lambda f: cond(t)(f)

ZERO = lambda f: lambda x: x
ADD1 = lambda n: (lambda f: lambda x: f(n (f)(x)))
ADD = lambda n: lambda m: n(ADD1)(m)
ONE = ADD1(ZERO)
IS_ZERO = lambda n: n(lambda x: FALSE)(TRUE)
SUB1 = (
  (lambda n:
   lambda f:
   lambda x: n (lambda g: lambda h: h(g(f)))
               (lambda u: x)
               (lambda u: u))
)
MULT = (
  lambda n: lambda m: n(lambda x:
                        ADD(x)(m)) (ZERO)
)

SIX = ADD1(ADD1(ADD1(ADD1(ADD1(ADD1(ZERO))))))

FACT = (

```

---

```python

print(
      (lambda myself: (lambda n: ((lambda cond: lambda t: lambda f: cond(t)(f))((lambda n: n(lambda x: lambda t: lambda f: f(lambda x:
      x))(lambda t: lambda f: t(lambda x: x)))(n))(lambda _: (lambda n: (lambda f: lambda x: f(n (f)(x))))(lambda f: lambda x:
      x))(lambda _: (lambda n: lambda m: n(lambda x:(lambda n: lambda m: n(lambda n: (lambda f: lambda x: f(n (f)(x))))(m))(x)(m))
      (lambda f: lambda x: x))(n)(myself(myself)(((lambda n:lambda f:lambda x: n (lambda g: lambda h: h(g(f)))(lambda u: x)(lambda u:
      u)))(n)))))))(lambda myself: (lambda n: ((lambda cond: lambda t: lambda f: cond(t)(f))((lambda n: n(lambda x: lambda t: lambda
      f: f(lambda x: x))(lambda t: lambda f: t(lambda x: x)))(n))(lambda _: (lambda n: (lambda f: lambda x: f(n (f)(x))))(lambda f:
      lambda x: x))(lambda _: (lambda n: lambda m: n(lambda x:(lambda n: lambda m: n(lambda n: (lambda f: lambda x: f(n
      (f)(x))))(m))(x)(m)) (lambda f: lambda x: x))(n)(myself(myself)(((lambda n:lambda f:lambda x: n (lambda g: lambda h:
      h(g(f)))(lambda u: x)(lambda u: u)))(n)))))))( (lambda n: (lambda f: lambda x: f(n (f)(x))))((lambda n: (lambda f: lambda
      x: f(n (f)(x))))((lambda n: (lambda f: lambda x: f(n (f)(x))))((lambda n: (lambda f: lambda x: f(n (f)(x))))((lambda n:
      (lambda f: lambda x: f(n (f)(x))))((lambda n: (lambda f: lambda x: f(n (f)(x))))(lambda f: lambda x: x)))))))

  (lambda x: x + 1)(0)
)

```

---

![left](http://media0.giphy.com/media/uTuLngvL9p0Xe/giphy.gif)

# [fit] 720

---

![](http://media4.giphy.com/media/l2JhtKtDWYNKdRpoA/giphy.gif)

> So why the f**k would you do this?
-- You, the audience member

---

# [fit] 1. It's *fun*

---

# [fit] 2. It gives *perspective*

---

Like isn't it cool how functions can represent the basic building blocks of programming

---

# [fit] 3. This is where *functional programming* stems from

^ C is to turing machines what Haskell is to lambda calculus

---

# [fit] 4. A simple alternative to *Turing Machines*

^ Turing machines are powerful but defining algorithms for it is like programming assembly. This is much more like a high level language.

---

^ We spend most of our time thinking about how to make great software. Lambda calculus is a simple abstraction over every model of computation. 

---

![](http://media0.giphy.com/media/xUPGcxpCV81ebKh7Vu/giphy.gif)

---

![Giphy](http://media2.giphy.com/media/3oriO5EMLxl1f7funu/giphy.gif)

---